@@ -1,4 +1,4 @@
-# Copyright (c) 2017-2018 Dell Inc. or its subsidiaries.
+# Copyright (c) 2020 Dell Inc. or its subsidiaries.
 # All Rights Reserved.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
@@ -17,6 +17,7 @@ import ast
 
 from oslo_log import log as logging
 
+from cinder import exception
 from cinder import interface
 from cinder.volume import driver
 from cinder.volume.drivers.dell_emc.powermax import common
@@ -122,21 +123,17 @@ class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
               - Switch to Unisphere REST API public replication endpoints
               - Support for multiple replication devices
               - Pools bug fix allowing 'None' variants (bug #1873253)
-        4.2.1 - Concurrent live migrations failure (bug #1875478)
-        4.2.2 - U4P failover lock not released on exception (#1875640)
-        4.2.3 - Live migrate remove rep vol from sg (bug #1875432)
-        4.2.4 - Rep validation & retype suspension fix (bug #1875433)
-        4.2.5 - Create vol suspend fix & DeviceID check (bug #1877976)
-        4.2.6 - Volume migrate exception handling (bug #1886662, #1874187)
-        4.2.7 - Fix to enable legacy volumes to live migrate (#1867163)
-        4.2.8 - Force flag to add to rep enable storage groups (#1892057)
-        4.2.9 - Failover group volume update (#1888831)
-        4.2.10 - Replica RDFG suspend fix (#1892718)
-        4.2.11 - Fix non-temporary snapshot delete (#1887962)
-        4.2.12 - Fix for create snapshot (#1939139)
+        4.3.0 - Changing from 91 to 92 REST endpoints
+              - Support for Port Group and Port load balancing
+                (bp powermax-port-load-balance)
+              - Fix to enable legacy volumes to live migrate (#1867163)
+              - Use of snap id instead of generation (bp powermax-snapset-ids)
+              - Support for Failover Abilities (bp/powermax-failover-abilities)
+        4.3.1 - Fix non-temporary snapshot delete (#1887962)
+        4.3.2 - Fix for create snapshot (#1939139)
     """
 
-    VERSION = "4.2.12"
+    VERSION = "4.3.2"
 
     # ThirdPartySystems wiki
     CI_WIKI_NAME = "EMC_VMAX_CI"
@@ -150,6 +147,8 @@ class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
             self.VERSION,
             configuration=self.configuration,
             active_backend_id=self.active_backend_id)
+        self.performance = self.common.performance
+        self.rest = self.common.rest
         self.zonemanager_lookup_service = fczm_utils.create_lookup_service()
 
     @classmethod
@@ -306,7 +305,7 @@ class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
         """
         device_number = device_info['hostlunid']
         target_wwns, init_targ_map = self._build_initiator_target_map(
-            volume, connector)
+            volume, connector, device_info)
 
         data = {'driver_volume_type': 'fibre_channel',
                 'data': {'target_lun': device_number,
@@ -462,11 +461,12 @@ class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
 
         return data
 
-    def _build_initiator_target_map(self, volume, connector):
+    def _build_initiator_target_map(self, volume, connector, device_info=None):
         """Build the target_wwns and the initiator target map.
 
         :param volume: the cinder volume object
         :param connector: the connector object
+        :param device_info: device_info
         :returns: target_wwns -- list, init_targ_map -- dict
         """
         target_wwns, init_targ_map = [], {}
@@ -475,6 +475,40 @@ class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
             self.common.get_target_wwns_from_masking_view(
                 volume, connector))
 
+        # If load balance is enabled we want to select only the FC target that
+        # has the lowest load of all ports in selected port group.
+        # Note: device_info in if condition as this method is called also for
+        # terminate connection, we only want to calculate load on initialise
+        # connection.
+        if device_info and self.performance.config.get('load_balance'):
+            try:
+                array_id = device_info.get('array')
+                masking_view = device_info.get('maskingview')
+                # Get PG from MV
+                port_group = self.rest.get_element_from_masking_view(
+                    array_id, masking_view, portgroup=True)
+                # Get port list from PG
+                port_list = self.rest.get_port_ids(array_id, port_group)
+                # Get lowest load port in PG
+                load, metric, port = self.performance.process_port_load(
+                    array_id, port_list)
+                LOG.info("Lowest %(met)s load port is %(port)s: %(load)s",
+                         {'met': metric, 'port': port, 'load': load})
+                # Get target WWN
+                port_details = self.rest.get_port(array_id, port)
+                port_info = port_details.get('symmetrixPort')
+                port_wwn = port_info.get('wwn_node')
+                LOG.info("Port %(p)s WWN: %(wwn)s",
+                         {'p': port, 'wwn': port_wwn})
+                # Set lowest load port WWN as FC target for connection
+                fc_targets = [port_wwn]
+            except exception.VolumeBackendAPIException:
+                LOG.error("There was an error calculating port load, "
+                          "reverting to default target selection.")
+                fc_targets, __ = (
+                    self.common.get_target_wwns_from_masking_view(
+                        volume, connector))
+
         if self.zonemanager_lookup_service:
             fc_targets.extend(metro_fc_targets)
             mapping = (
@@ -501,18 +535,7 @@ class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
         """
         self.common.extend_volume(volume, new_size)
 
-    def get_volume_stats(self, refresh=False):
-        """Get volume stats.
-
-        :param refresh: boolean -- If True, run update the stats first.
-        :returns: dict -- the stats dict
-        """
-        if refresh:
-            self.update_volume_stats()
-
-        return self._stats
-
-    def update_volume_stats(self):
+    def _update_volume_stats(self):
         """Retrieve stats info from volume group."""
         LOG.debug("Updating volume stats")
         data = self.common.update_volume_stats()
